# -*- coding: utf-8 -*-

import logging, time, re
import pprint, threading
from constants import DEBUG_CONFIG
from flask import request, json
from functools import wraps
import datetime
from logging.handlers import TimedRotatingFileHandler
import sys, traceback
#from timed_ompressed_rotating_file_handler import TimedCompressedRotatingFileHandler

def get_logger(name):

    logger_file_name = '/var/log/amc/' + str(name) + '.log'
    formatter = logging.Formatter('\n%(asctime)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    if logger.handlers:
        logger.handlers = []
    #fh = logging.handlers.RotatingFileHandler(logger_file_name, maxBytes=50000000, backupCount=2)
    #fh = TimedCompressedRotatingFileHandler(logger_file_name, when="midnight", backupCount=2)
    fh = TimedRotatingFileHandler(logger_file_name)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    return logger

class _Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Singleton(_Singleton('SingletonMeta', (object,), {})): pass


class Debugger(Singleton):

    def __init__(self):

        self._logger = None
        self._debugging = False
        self._debug_address = None
        self._debug_user = None
        self._cluster_id = None
        self._cluster_seed = None
        self._cluster_node_list = None
        self._start_time = None
        self._debug_duration = DEBUG_CONFIG['inactivity_timeout']

    def start(self, user_address, cluster_id, seed_node, node_list, debug_duration):
        if not self._debugging:
            if self._logger is not None:
                self.print_terminate(True)
                self._logger = None

            self._debug_address = user_address
            self._cluster_id = cluster_id
            self._cluster_seed = seed_node
            self._cluster_node_list = node_list
            self._logger = get_logger(DEBUG_CONFIG['filename'])
            self._debugging = True
            self._start_time = int(round(time.time()*1000))
            self._debug_duration = debug_duration * 60000
            return self.print_initiate()
        else:
            return False


    def stop(self, user_requested = True):
        requestOrigin = self.is_request_from_initiator()

        if self._debugging and requestOrigin:
            if self._logger is not None:
                self.print_terminate(user_requested)
                self._logger = None

            self._debugging = False
            self._debug_address = None
            self._debug_user = None
            self._cluster_id = None
            self._cluster_seed = None
            self._cluster_node_list = None

        if not self._debugging:
            return True

        return False


    def restart(self, user_address, cluster_id, seed_node, node_list, debug_duration):
        requestOrigin = self.is_request_from_initiator()

        if requestOrigin:
            if self._debugging:
                self.stop()

            if self._debugging:
                return False

            return self.start(user_address, cluster_id, seed_node, node_list, debug_duration)

        return False


    def status(self):

        status = {}
        status['debugging'] = ("ON" if self._debugging else "OFF")
        current_time = int(round(time.time()*1000))

        if self._debugging:
            status['initiator'] = self._debug_address
            status['start_time'] = self._start_time
            if current_time - self._start_time >= self._debug_duration:
                self.stop()
        else:
            status['initiator'] = None
        return status


    def print_terminate(self, user_requested):
        if user_requested:
            return self.print_content(('DEBUG TEMINATED' + ((' BY '+ str(self._debug_user)) if self._debug_user is not None else '') +' FROM '+ str(self._debug_address)), None)
        else:
            return self.print_content('DEBUG TEMINATED', None)


    def print_initiate(self):

        self.print_content(('DEBUG INITIATED' + ((' BY '+ str(self._debug_user)) if self._debug_user is not None else '') +' FROM '+ str(self._debug_address)), None)
        return self.print_content("INITIATION INFO", ("\nCluster Security : " +
                ("Enabled" if self._debug_user is not None else "Disabled") +
                "," + "Cluster ID : " + str(self._cluster_id) +
                "," + "Cluster Seed : " + str(self._cluster_seed) +
                "," + "Cluster Nodes : " + str(self._cluster_node_list)))


    def print_reset(self):

        return self.print_content(('DEBUG RESTARTED' + (('BY '+ str(self._debug_user)) if self._debug_user is not None else '') +' FROM '+ str(self._debug_address)), None)


    def print_content(self, title, content):
        stripedContent = None
        printable = None

        if content is not None:
            stripedContent = re.sub(r"\s+", "", self.prettify(content))
            stripedContent = re.sub("nextLine", "\n", stripedContent)
            stripedContent = re.sub("newTab", "\t", stripedContent)

        if stripedContent is not None:
            printable = self.get_decorated_content(title) + ": " + stripedContent
        else:
            printable = self.get_decorated_content(title)

        self._logger.info(printable)

        return True

    def get_decorated_content(self, content):
        content = '<<<< ' + str(content) +' >>>>'
        return content

    def is_request_from_initiator(self):
        try:
            if str(request.remote_addr) == self._debug_address:
                return True
            return False
        except RuntimeError as err:
            return str(err)

    def is_cluster_id_set_from_initiator(self, args):
        try:

            cluster_id = args['cluster']

            if ((str(type(cluster_id))).find("class") != -1):
                cluster_id = cluster_id.get_cluster_id()

            if self._cluster_id == cluster_id:
                return True

        except KeyError:
            return True

        return False


    def prettify(self, content):
        result = ""

        if ((str(type(content))).find("class") != -1):
            result = "{'object__type__': " + pprint.pformat(content) + ",'object__dict__': " + pprint.pformat(content.__dict__) + "}"

        elif (str(type(content))).find("dict") != -1 :
            dict_empty = True
            result = "{"
            for key in content:
                result += "'" + str(key) + "': " + self.prettify(content[key]) + ","
                dict_empty = False

            if not dict_empty:
                result = result[:-1]

            result += "}"

        elif (str(type(content))).find("set") != -1 :
            pprint.pformat(content)

        elif (str(type(content))).find("str") != -1 :
            result = content

        else:
            result = json.dumps(content)

        return str(result)


    def check_and_log(self, fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            #Check if we are working inside flask's request context
            requestOrigin = self.is_request_from_initiator()
            clusterOrigin = self.is_cluster_id_set_from_initiator(kwargs)
            current_time = int(round(time.time()*1000))
            response = None
            try:
                if  ((self._debugging) and
                    ((DEBUG_CONFIG['allow_request'] == 'allow_debug_user' and requestOrigin) or DEBUG_CONFIG['allow_request'] == 'allow_all' ) and
                    (((DEBUG_CONFIG['cluster_level'] == 'requested' and clusterOrigin) or type(requestOrigin) == str) or DEBUG_CONFIG['cluster_level'] == 'all' )):

                    if current_time - self._start_time >= self._debug_duration:
                        self.stop()

                    if requestOrigin == True:
                        context = "HTTP Req"
                    else:
                        context = "Poll Cycle"

                    requestStartTime = int(round(time.time()*1000000))
                    response = fn(*args, **kwargs)
                    requestEndTime = int(round(time.time()*1000000))

                    requestHead = "ID TAG | " + str( threading.current_thread().name ) + " | " + (fn.__wrappedmodule__ if hasattr(fn, '__wrappedmodule__') else fn.__module__) + "@" + (fn.__wrappedname__ if hasattr(fn, '__wrappedname__') else fn.__name__) + " | " + context + " | " + "Execution time[ " + str(requestEndTime - requestStartTime) + "]"

                    requestBody = ""
                    #Print flask request object essential attributes if available
                    if requestOrigin == True:
                        requestBody = requestBody + "nextLine Request Context : newTab" + self.prettify({'url' : (request.script_root + request.path), 'method' : ('['+ request.method +']'), 'arguments': dict(request.form.items() + request.args.items())})

                    # if DEBUG_CONFIG['log_call_params'] and (DEBUG_CONFIG['log_level'] == 'end_points' or DEBUG_CONFIG['log_level'] == 'trace' or DEBUG_CONFIG['log_level'] == 'user_facing'):
                    # requestBody = requestBody + "nextLine Params : newTab" + self.prettify(kwargs)

                    if DEBUG_CONFIG['log_level'] == 'end_points' or DEBUG_CONFIG['log_level'] == 'trace' or DEBUG_CONFIG['log_level'] == 'user_facing':
                        requestBody = requestBody + "nextLine Response : newTab" + self.prettify(response)

                    self.print_content(requestHead, requestBody)
                else:
                    response = fn(*args, **kwargs)
            except Exception:
                if self._debugging:
                    sys.stderr.write("ERROR in debugger run")
                    self.stop(False)
                    sys.stderr.write((traceback.format_exc()))
                if response is None:
                    response = fn(*args, **kwargs)

            return response

        return decorator

